"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const definition_1 = require("graphql/type/definition");
const directives_1 = require("graphql/type/directives");
const introspection_1 = require("graphql/type/introspection");
const find_1 = require("graphql/jsutils/find");
const invariant_1 = require("graphql/jsutils/invariant");
const objectValues_1 = require("graphql/jsutils/objectValues");
const GraphQLDeferDirective_1 = require("./GraphQLDeferDirective");
class GraphQLSchema {
    constructor(config) {
        if (config && config.assumeValid) {
            this.__validationErrors = [];
        }
        else {
            invariant_1.default(typeof config === 'object', 'Must provide configuration object.');
            invariant_1.default(!config.types || Array.isArray(config.types), `"types" must be Array if provided but got: ${String(config.types)}.`);
            invariant_1.default(!config.directives || Array.isArray(config.directives), '"directives" must be Array if provided but got: ' +
                `${String(config.directives)}.`);
            invariant_1.default(!config.allowedLegacyNames || Array.isArray(config.allowedLegacyNames), '"allowedLegacyNames" must be Array if provided but got: ' +
                `${String(config.allowedLegacyNames)}.`);
        }
        this.__allowedLegacyNames = config.allowedLegacyNames;
        this._queryType = config.query;
        this._mutationType = config.mutation;
        this._subscriptionType = config.subscription;
        this._directives = config.directives || [
            directives_1.GraphQLIncludeDirective,
            directives_1.GraphQLSkipDirective,
            directives_1.GraphQLDeprecatedDirective,
            GraphQLDeferDirective_1.default,
        ];
        this.astNode = config.astNode;
        let initialTypes = [
            this.getQueryType(),
            this.getMutationType(),
            this.getSubscriptionType(),
            introspection_1.__Schema,
        ];
        const types = config.types;
        if (types) {
            initialTypes = initialTypes.concat(types);
        }
        let typeMap = Object.create(null);
        typeMap = initialTypes.reduce(typeMapReducer, typeMap);
        typeMap = this._directives.reduce(typeMapDirectiveReducer, typeMap);
        this._typeMap = typeMap;
        this._implementations = Object.create(null);
        Object.keys(this._typeMap).forEach(typeName => {
            const type = this._typeMap[typeName];
            if (definition_1.isObjectType(type)) {
                type.getInterfaces().forEach(iface => {
                    if (definition_1.isInterfaceType(iface)) {
                        const impls = this._implementations[iface.name];
                        if (impls) {
                            impls.push(type);
                        }
                        else {
                            this._implementations[iface.name] = [type];
                        }
                    }
                });
            }
        });
    }
    getQueryType() {
        return this._queryType;
    }
    getMutationType() {
        return this._mutationType;
    }
    getSubscriptionType() {
        return this._subscriptionType;
    }
    getTypeMap() {
        return this._typeMap;
    }
    getType(name) {
        return this.getTypeMap()[name];
    }
    getPossibleTypes(abstractType) {
        if (definition_1.isUnionType(abstractType)) {
            return abstractType.getTypes();
        }
        return this._implementations[abstractType.name];
    }
    isPossibleType(abstractType, possibleType) {
        let possibleTypeMap = this._possibleTypeMap;
        if (!possibleTypeMap) {
            this._possibleTypeMap = possibleTypeMap = Object.create(null);
        }
        if (!possibleTypeMap[abstractType.name]) {
            const possibleTypes = this.getPossibleTypes(abstractType);
            invariant_1.default(Array.isArray(possibleTypes), `Could not find possible implementing types for ${abstractType.name} ` +
                'in schema. Check that schema.types is defined and is an array of ' +
                'all possible types in the schema.');
            possibleTypeMap[abstractType.name] = possibleTypes.reduce((map, type) => ((map[type.name] = true), map), Object.create(null));
        }
        return Boolean(possibleTypeMap[abstractType.name][possibleType.name]);
    }
    getDirectives() {
        return this._directives;
    }
    getDirective(name) {
        return find_1.default(this.getDirectives(), directive => directive.name === name);
    }
}
exports.GraphQLSchema = GraphQLSchema;
;
function typeMapReducer(map, type) {
    if (!type) {
        return map;
    }
    if (definition_1.isWrappingType(type)) {
        return typeMapReducer(map, type.ofType);
    }
    if (map[type.name]) {
        invariant_1.default(map[type.name] === type, 'Schema must contain unique named types but contains multiple ' +
            `types named "${type.name}".`);
        return map;
    }
    map[type.name] = type;
    let reducedMap = map;
    if (definition_1.isUnionType(type)) {
        reducedMap = type.getTypes().reduce(typeMapReducer, reducedMap);
    }
    if (definition_1.isObjectType(type)) {
        reducedMap = type.getInterfaces().reduce(typeMapReducer, reducedMap);
    }
    if (definition_1.isObjectType(type) || definition_1.isInterfaceType(type)) {
        objectValues_1.default(type.getFields()).forEach(field => {
            if (field.args) {
                const fieldArgTypes = field.args.map(arg => arg.type);
                reducedMap = fieldArgTypes.reduce(typeMapReducer, reducedMap);
            }
            reducedMap = typeMapReducer(reducedMap, field.type);
        });
    }
    if (definition_1.isInputObjectType(type)) {
        objectValues_1.default(type.getFields()).forEach(field => {
            reducedMap = typeMapReducer(reducedMap, field.type);
        });
    }
    return reducedMap;
}
function typeMapDirectiveReducer(map, directive) {
    if (!directives_1.isDirective(directive)) {
        return map;
    }
    return directive.args.reduce((_map, arg) => typeMapReducer(_map, arg.type), map);
}
//# sourceMappingURL=GraphQLSchema.js.map