"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const type_1 = require("graphql/type");
const starWarsData_1 = require("./starWarsData");
const GraphQLDeferDirective_1 = require("../GraphQLDeferDirective");
const episodeEnum = new type_1.GraphQLEnumType({
    name: 'Episode',
    description: 'One of the films in the Star Wars Trilogy',
    values: {
        NEWHOPE: {
            value: 4,
            description: 'Released in 1977.',
        },
        EMPIRE: {
            value: 5,
            description: 'Released in 1980.',
        },
        JEDI: {
            value: 6,
            description: 'Released in 1983.',
        },
    },
});
const characterInterface = new type_1.GraphQLInterfaceType({
    name: 'Character',
    description: 'A character in the Star Wars Trilogy',
    fields: () => ({
        id: {
            type: new type_1.GraphQLNonNull(type_1.GraphQLString),
            description: 'The id of the character.',
        },
        name: {
            type: type_1.GraphQLString,
            description: 'The name of the character.',
        },
        friends: {
            type: new type_1.GraphQLList(characterInterface),
            description: 'The friends of the character, or an empty list if they ' +
                'have none.',
        },
        appearsIn: {
            type: new type_1.GraphQLList(episodeEnum),
            description: 'Which movies they appear in.',
        },
        secretBackstory: {
            type: type_1.GraphQLString,
            description: 'All secrets about their past.',
        },
    }),
    resolveType(character) {
        if (character.type === 'Human') {
            return humanType;
        }
        if (character.type === 'Droid') {
            return droidType;
        }
    },
});
const humanType = new type_1.GraphQLObjectType({
    name: 'Human',
    description: 'A humanoid creature in the Star Wars universe.',
    fields: () => ({
        id: {
            type: new type_1.GraphQLNonNull(type_1.GraphQLString),
            description: 'The id of the human.',
        },
        name: {
            type: type_1.GraphQLString,
            description: 'The name of the human.',
        },
        nonNullField: {
            type: new type_1.GraphQLNonNull(type_1.GraphQLString),
            description: 'This field cannot be null.',
        },
        friends: {
            type: new type_1.GraphQLList(characterInterface),
            description: 'The friends of the human, or an empty list if they have none.',
            resolve: human => starWarsData_1.getFriends(human),
        },
        appearsIn: {
            type: new type_1.GraphQLList(episodeEnum),
            description: 'Which movies they appear in.',
        },
        homePlanet: {
            type: type_1.GraphQLString,
            description: 'The home planet of the human, or null if unknown.',
        },
        secretBackstory: {
            type: type_1.GraphQLString,
            description: 'Where are they from and how they came to be who they are.',
            resolve() {
                throw new Error('secretBackstory is secret.');
            },
        },
        soulmate: {
            type: new type_1.GraphQLNonNull(characterInterface),
            description: 'Everyone has a soulmate and should error otherwise.',
        },
        weapon: {
            type: new type_1.GraphQLObjectType({
                name: 'Weapon',
                fields: {
                    name: {
                        type: type_1.GraphQLString,
                        description: 'Name of the weapon',
                    },
                    strength: {
                        type: type_1.GraphQLString,
                        description: 'Strength of weapon',
                    },
                },
            }),
        },
    }),
    interfaces: [characterInterface],
});
const droidType = new type_1.GraphQLObjectType({
    name: 'Droid',
    description: 'A mechanical creature in the Star Wars universe.',
    fields: () => ({
        id: {
            type: new type_1.GraphQLNonNull(type_1.GraphQLString),
            description: 'The id of the droid.',
        },
        name: {
            type: type_1.GraphQLString,
            description: 'The name of the droid.',
        },
        friends: {
            type: new type_1.GraphQLList(characterInterface),
            description: 'The friends of the droid, or an empty list if they have none.',
            resolve: droid => starWarsData_1.getFriends(droid),
        },
        appearsIn: {
            type: new type_1.GraphQLList(episodeEnum),
            description: 'Which movies they appear in.',
        },
        secretBackstory: {
            type: type_1.GraphQLString,
            description: 'Construction date and the name of the designer.',
            resolve() {
                throw new Error('secretBackstory is secret.');
            },
        },
        primaryFunction: {
            type: type_1.GraphQLString,
            description: 'The primary function of the droid.',
        },
    }),
    interfaces: [characterInterface],
});
const queryType = new type_1.GraphQLObjectType({
    name: 'Query',
    fields: () => ({
        hero: {
            type: characterInterface,
            args: {
                episode: {
                    description: 'If omitted, returns the hero of the whole saga. If ' +
                        'provided, returns the hero of that particular episode.',
                    type: episodeEnum,
                },
            },
            resolve: (root, { episode }) => starWarsData_1.getHero(episode),
        },
        human: {
            type: humanType,
            args: {
                id: {
                    description: 'id of the human',
                    type: new type_1.GraphQLNonNull(type_1.GraphQLString),
                },
            },
            resolve: (root, { id }) => starWarsData_1.getHuman(id),
        },
        droid: {
            type: droidType,
            args: {
                id: {
                    description: 'id of the droid',
                    type: new type_1.GraphQLNonNull(type_1.GraphQLString),
                },
            },
            resolve: (root, { id }) => starWarsData_1.getDroid(id),
        },
    }),
});
exports.StarWarsSchema = new type_1.GraphQLSchema({
    query: queryType,
    types: [humanType, droidType],
    directives: [GraphQLDeferDirective_1.default],
});
//# sourceMappingURL=starWarsSchema.js.map